# -*- coding: utf-8 -*-
""" Configuration manager. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
from os.path import exists

# 3rd party imports
import yaml


class Config(object):
    """ Config loader class.

    Best used by subclassing and using declarative approach. There are few class
    global fields that can be set to customize the loader

    :VERSION:
        This defines the config version parsed by the iteration. Only change
        this if you change the config structure for your project. If the version
        stored in the config file and this setting differ the loading will raise
        exception about version mismatch.
    :defaults:
        A dict with default values for the configuration. Each key is the
        config value name and each value is the default value to use. Those will
        only remain in the config if they are not defined in any other way. If
        the value is a callable it will be called with the current config state
        as it's only argument. The return value will be used as the config
        default.
    :deserializers:
        Sometimes it's easier to store a config value in a python specific type.
        Since we can't store python data types directly in the files we can
        store a representation of them. This is a map where each key is the
        config value name and the value is a function that will receive the
        value loaded and then return what should be stored in the config. This
        can be used for example to convert an ISO date string to a `datetime`
        object and store that in the config instead.
    """
    VERSION = 1
    defaults = {}
    deserializers = {}
    resolution = []

    class Error(RuntimeError):
        """ Base class for all config errors. """
        pass

    def __init__(self, conf_path=None):
        self.conf_path = conf_path or self._discover_config()
        self.config = {}

    def load(self, required=False):
        """ Load the configuration. """
        if self.conf_path is not None:
            config = self._read()
        elif not required:
            config = {'version': self.VERSION}
        else:
            raise RuntimeError(
                "No configuration found. Please provide one of:\n" +
                "\n".join(self.resolution)
            )

        self.validate(config)
        self.deserialize(config)
        self.apply_defaults(config)

        self.config = config
        return config

    def __getattr__(self, item):
        """

        :param str item:
        :return:
        """
        if item.isupper():
            return self.config[item]

        return super(Config, self).__getattr__(item)

    def validate(self, config):
        """ Validate config

        :param Dict[str, Any] config:
            Config as loaded from external sources.
        :raises Config.Error:
            Raised if the config can't be validated.
        """
        if 'version' not in config:
            raise Config.Error("Unknown config version: {}".format(
                self.conf_path
            ))

        if config['version'] != self.VERSION:
            msg = "Unsupported config version: {}. Should be {}".format(
                config['version'], self.VERSION
            )
            raise Config.Error(msg)

    def deserialize(self, config):
        """ Deserialize config.

        This will convert all the primitive values to the appropriate python
        data structures (like date string to date).

        :param Dict[str, Any] config:
            The config to deserialize values in.
        """
        for name, deserialize in self.deserializers.items():
            if name in config:
                config[name] = deserialize(config[name])

    def apply_defaults(self, config):
        """ Apply defaults to the configuration.

        This step is called at the end  of config loading process (after
        deserialization) so you can use the existing config values as a basis
        for your defaults. Here you can also create new complex variables that
        are dependent on other variables.

        :param Dict[str, Any] config:
            The config dict where we apply the defaults.
        """
        for name, value in self.defaults.items():
            if callable(value):
                value = value(config)
            config.setdefault(name, value)

    def _discover_config(self):
        """ Find configuration file.

        The config resolution is as follows:
        - ./scaffyhub.yaml
        - /etc/scaffyhub.yaml
        - <scaffyhub package>/run/config.yaml
        """
        return next((x for x in self.resolution if exists(x)), None)

    def _read(self):
        try:
            # app_cfg = json.load(open(conf_path))
            with open(self.conf_path) as fp:
                config = yaml.load(fp.read())

        except yaml.YAMLError as ex:
            raise RuntimeError("Failed to load config {}: {}".format(
                self.conf_path,
                str(ex)
            ))

        return config



# -*- coding: utf-8 -*-
""" API client for scaffyhub. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
from datetime import datetime
from functools import wraps
from os.path import exists

# 3rd party imports
import requests

# local imports
from scaffy.config import settings
from scaffy.core import shell


def print_api_errors(fn):
    """ A handy decorator to print out all uncaught ApiErrors. """
    @wraps(fn)
    def wrapper(*args, **kw):   # pylint: disable=missing-docstring
        try:
            return fn(*args, **kw)
        except ApiError as ex:
            shell.cprint('^31{}^0'.format(str(ex)))

    return wrapper


class ApiError(RuntimeError):
    """ Base class for all API client. errors. """
    def __init__(self, response, message=None):
        if message is None:
            try:
                message = '[{code}] {detail}'.format(
                    code=response.status_code,
                    detail=json.dumps(response.json()['detail'], indent=2)
                )
            except Exception:
                message = '[{code}] Unknown error:\n{body}'.format(
                    code=response.status_code,
                    body=response.text
                )

        super(ApiError, self).__init__(message)
        self.response = response


class ApiClientSectionBase(object):
    """ Base class for API sections.

    To avoid having a huge client class with all possible API calls the API
    client is divided into sections. The sections are setup/initialized in the
    `Client._init_api_sections()` method and if you create new ones you should
    subclass this class and add a new section there.
    """
    def __init__(self, client):
        self.api = client


class Client(object):
    """ scaffyhub API client. """

    class BackendNotFound(RuntimeError):
        """ Raised when the client can't connect to the backend."""
        pass

    def __init__(self, backend_url=None):
        self.backend_url = backend_url or self._discover_backend()
        self._token = None
        self._init_api_sections()

    def _discover_backend(self):
        backends = settings.BACKENDS

        for backend in backends:
            if backend.endswith('/'):
                backend = backend[:-1]

            try:
                url = backend + '/api/ping'
                r = requests.get(url)
                if r.status_code == 200:
                    return backend
            except:
                pass
        raise Client.BackendNotFound("Backend unreachable: {}".format(backends))

    def _init_api_sections(self):
        """ Initialize API section.

        For more information, see `ApiClientSectionBase`
        """
        from .scaffold import ScaffoldApi

        self.scaffold = ScaffoldApi(self)

    @property
    def token(self):
        """ Get the authorization token.

        If it doesn't have a cached value it will try to read it from
        `settings.CREDENTIALS_PATH`.
        """
        if self._token is None:
            if exists(settings.CREDENTIALS_PATH):
                with open(settings.CREDENTIALS_PATH) as fp:
                    self._token = fp.read()

        return self._token

    @token.setter
    def token(self, token):
        self._token = token

        with open(settings.CREDENTIALS_PATH, 'w') as fp:
            fp.write(token)

    def url(self, relative_url):
        """ Get the full URL from the API relative url."""
        if not relative_url.startswith('/'):
            relative_url = '/' + relative_url

        return self.backend_url + relative_url

    def login(self, username, password):
        """ Login to scaffyhub

        On successful login it will store the authorization token for later use.

        :param str username:
            scaffyhub username
        :param str password:
            scaffyhub password
        """
        r = requests.post(self.url('/api/user/login'), json={
            'username': username,
            'password': password
        })

        if r.status_code != 200:
            raise ApiError(r)

        self.token = r.json()['jwt_token']

    def get(self, url, params=None, headers=None, **kw):
        """ Proxy for requests.get that handles auth. """
        if self.token is not None:
            headers = headers or {}
            headers['Authorization'] = 'JWT ' + self.token

        return requests.get(self.url(url), params, headers=headers, **kw)

    def post(self, url, data=None, headers=None, **kw):
        """ Proxy for requests.post that handles auth. """
        if self.token is not None:
            headers = headers or {}
            headers['Authorization'] = 'JWT ' + self.token

        return requests.post(self.url(url), json=data, headers=headers, **kw)

    def put(self, url, data=None, headers=None, **kw):
        """ Proxy for requests.put that handles auth. """
        if self.token is not None:
            headers = headers or {}
            headers['Authorization'] = 'JWT ' + self.token

        return requests.put(self.url(url), json=data, headers=headers, **kw)

    def delete(self, url, headers=None, **kw):
        """ Proxy for requests.delete that handles auth. """
        if self.token is not None:
            headers = headers or {}
            headers['Authorization'] = 'JWT ' + self.token

        return requests.delete(self.url(url), headers=headers, **kw)



# -*- coding: utf-8 -*-
""" Commands for interacting with the remote service. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import json
import os
import sys

# 3rd party imports
import click

# local imports
from scaffy import remote
from scaffy import scaffold
from scaffy.core import log
from scaffy.core import shell
from scaffy.scaffold import Scaffold
from . import cli


@cli.group('hub')
def hub_cli():
    """ ScaffyHub commands. """
    pass


@hub_cli.command()
@click.option(
    '-u', '--username',
    type=str,
    prompt='Username',
    help='scaffyhub username'
)
@click.option(
    '-p', '--password',
    type=str,
    prompt='Password',
    hide_input=True,
    help='scaffyhub password'
)
@remote.print_api_errors
def login(username, password):
    """ Login to scaffyhub.

    After login the authorization token will be stored in scaffy ~/.config
    directory for future use. The username/password is not stored anywhere and
    once the token expires the user will log again. The app supports rolling
    session so every time you interact with scaffyhub you will get the new token
    and your session time will be extended.
    """
    api = remote.Client()

    log.info("Authenticating with {}".format(api.backend_url))
    api.login(username, password)
    log.info("Logged in")


@hub_cli.command()
@click.argument('name', type=str)
@remote.print_api_errors
def push(name):
    """ Push scaffold to scaffyhub.

    The scaffold file will be uploaded to scaffyhub. You can then pull that
    scaffold from any machine using `scaffy pull`.
    """
    local_store = scaffold.LocalStore()
    api = remote.Client()

    item = local_store.load(name)

    # Upload the scaffold
    log.info("Uploading ^35{} ^32to ^34{}".format(item.name, api.backend_url))
    try:
        api.scaffold.upload(item)
        log.info("Done")
        # shell.cprint('^33[{}]: ^90{}', 200, json.dumps(
        #     remote_item.serialize(), indent=2
        # ))
    except remote.ApiError as ex:
        log.err("[{status}] {detail}:\n{body}".format(
            status=ex.response.status_code,
            detail=str(ex),
            body=json.dumps(ex.response.json(), indent=2)
        ))


@hub_cli.command()
@click.argument('name', type=str)
@remote.print_api_errors
def pull(name):
    """ Pull a scaffold with the given name from scaffyhub. """
    local_store = scaffold.LocalStore()
    api = remote.Client()

    remote_item = api.scaffold.get_by_name(name)

    if remote_item is None:
        log.err("'{}' does not exist".format(name))
        sys.exit(-1)

    temp_path = remote_item.download()
    item = Scaffold.load_from_file(temp_path)

    # This will physically write the scaffold file to a local store, thus
    # making a copy so the original can be deleted.
    local_store.add(item)
    os.remove(temp_path)


@hub_cli.command('list')
@remote.print_api_errors
def hub_list():
    """ List scaffolds locally on the hub. """
    api = remote.Client()

    shell.cprint("^32Remote:\n^0")
    for item in api.scaffold.list():
        shell.cprint("  ^90{}  ^0{}", item.created, item.name)

    shell.cprint('')


@hub_cli.command('delete')
@click.argument('names', nargs=-1, type=str, required=True)
def hub_delete(names):
    """ Delete selected scaffolds from the hub.

    Calling with multiple names will delete all of them with one command.
    """
    api = remote.Client()

    for remote_item in api.scaffold.list():
        if remote_item.name in names:
            try:
                remote_item.delete()
            except remote.ApiError as ex:
                log.err("[{status}] {detail}:\n{body}".format(
                    status=ex.response.status_code,
                    detail=str(ex),
                    body=ex.response.text
                ))



# -*- coding: utf-8 -*-
""" Basic commands.
TODO:
    - Add optional fields.
        - On creating you specify pairs of name_in_src=field_name, The marker
          must include the field name somehow.
        - On applying during the read, once a marker is encountered the first
          time, the user us prompted for the value. Then that value is used
          for all subsequent occurrences.
"""
from __future__ import absolute_import, unicode_literals

import json

# 3rd party imports
import click

# local imports
from scaffy import remote
from scaffy import scaffold
from scaffy.core import log
from scaffy.core import shell
from . import cli


@cli.command('info')
@click.argument('name', type=str, required=True)
@click.option(
    '-f', '--files', 'show_files',
    is_flag=True,
    help='Show the list of files in the scaffold'
)
@click.option(
    '-c', '--config', 'show_config',
    is_flag=True,
    help='Show scaffold config'
)
def info(name, show_files, show_config):
    """ Show details about a given scaffold

    Example:

        \033[1mscaffy info -cf my-scaffold \033[0m

    This will show the full information about 'my-scaffold' scaffold. This
    includes the scaffold config as well as a list of scaffold files.
    """
    store = scaffold.LocalStore()
    item = store.load(name)

    log.info("Name:     ^33{}", item.name)
    log.info("Size:     ^33{} kb", round(item.size / 1024))
    log.info("Created:  ^33{}", item.pretty_created)

    if show_config:
        shell.cprint("Config: {")
        for key, value in item.json_config.items():
            shell.cprint('  "{name}": ^33{value}^0'.format(
                name=key, value=json.dumps(value, indent=2)
            ))

        shell.cprint("}")

    if show_files:
        log.info("Files:")
        for path in item.files:
            if path in item.marked_files:
                log.info('  ^0{}', path)
            else:
                log.info('  ^90{}', path)


@cli.command('delete')
@click.argument('names', nargs=-1, type=str, required=True)
@remote.print_api_errors
def delete(names):
    """ Delete selected scaffolds from the local store.

    Calling with multiple names will delete all of them with one command.
    """
    for name in names:
        log.info("Deleting ^35{}".format(name))
        store = scaffold.LocalStore()
        store.delete(name)


@cli.command('list')
def list_scaffolds():
    """ List scaffolds in local store. """
    local = scaffold.LocalStore()

    shell.cprint("^32Local:\n^0")
    for item in local.scaffolds:
        shell.cprint("  ^90{}  ^0{}", item.created, item.name)

    shell.cprint('')



# -*- coding: utf-8 -*-
""" Scaffold API client. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
import tempfile

# 3rd party imports
import attr

# local imports
from . import ApiError, ApiClientSectionBase
from . import util


class ScaffoldApi(ApiClientSectionBase):
    """ Handle interaction with scaffold API.

    This should never  be created and only accessed through `Client` instance as
    `Client().scaffold`.
    """
    def list(self):
        """ Return a list of all scaffolds stored in the hub by the user.

        :return List[RemoteScaffold]:
            List of scaffolds stored on scaffyhub.
        """
        r = self.api.get('/api/scaffold')
        if r.status_code != 200:
            raise ApiError(r, "Failed to get list of scaffolds: {}".format(
                r.json()['detail']
            ))

        return [RemoteScaffold.from_dict(self.api, x) for x in r.json()]

    def get_by_name(self, name):
        """ Get scaffold metadata by name.

        :param str name:
            The name of the scaffold.
        :return RemoteScaffold:
            The remote scaffold metadata. You can download the actual scaffold
            file using `RemoteScaffold.download()`
        """
        for item in self.list():
            if item.name == name:
                return item

        return None

    def upload(self, scaffold):
        """ Upload the scaffold to scaffyhub.

        :param scaffy.scaffold.Scaffold scaffold:
            The scaffold to upload
        :return RemoteScaffold):
            A `RemoteScaffold` instance that represents the uploaded scaffold.
        """
        if scaffold.path is None:
            raise NotImplementedError(
                "No support for scaffolds not loaded from a file"
            )

        with open(scaffold.path, 'rb') as fp:
            r = self.api.post('/api/scaffold', files={'file': fp})

        if r.status_code != 201:
            raise ApiError(r, "Failed to push {}".format(scaffold.name))

        return RemoteScaffold.from_dict(self.api, r.json())


@attr.s
class RemoteScaffold(object):
    """ Represents a scaffold stored on scaffyhub.

    You should never need to create instances of this class manually and
    only receive them as results of scaffold API calls.
    """
    api = attr.ib()

    id = attr.ib(default=None)
    name = attr.ib(default=None)
    created = attr.ib(default=None, convert=util.parse_date)
    size = attr.ib(default=None)
    config = attr.ib(default=None)
    owner_id = attr.ib(default=None)
    sha1 = attr.ib(default=None)
    filename = attr.ib(default=None)

    @classmethod
    def from_dict(cls, client, values):
        """ Create new instance from a dictionary of values.

        :param scaffy.remote.Client client:
            API client instance. This is used to issue all API calls related
            to this instance (like download).
        :param dict values:
            A dict of values to use for the new instance. This is most likely
            returned by an API call.
        :return RemoteScaffold:
            A `RemoteScaffold` instance initialized with the given *values*.
        """
        values['api'] = client
        return RemoteScaffold(**values)

    def download(self, out_path=None, buff_size=16 * 1024):
        """

        :param str out_path:
            Path where the file will be saved. If not given, the scaffold will
            be downloaded into a temporary file.
        :param int buff_size:
            Read buffer size. This is the size of the buffer used to read the
            scaffyhub response.
        :return str:
            Path to the downloaded scaffold file. If no *out_path* was given,
            then this will be a path to a temporary file. The caller is
            responsible for deleting it (or leaving it to the OS).
        """
        r = self.api.get(
            '/api/scaffold/{}/download'.format(self.id),
            stream=True
        )

        if r.status_code != 200:
            raise ApiError(r, "Failed to download {}".format(self.name))

        if out_path is not None:
            fp = open(out_path, 'wb')
        else:
            fd, out_path = tempfile.mkstemp()
            fp = os.fdopen(fd, 'wb')

        for chunk in r.iter_content(chunk_size=buff_size):
            fp.write(chunk)

        fp.close()

        return out_path

    def delete(self):
        """
        Delete ScaffyHub item represented by this instance.

        :raises ApiError:
            Raised when there is any problem with the API call, including an
            unexpected status code.
        """
        r = self.api.delete('/api/scaffold/{}'.format(self.id))

        if r.status_code != 204:
            raise ApiError(r, "Failed to delete {}".format(self.name))



# -*- coding: utf-8 -*-
""" Commands package.

All commands (and only commands) should be defined inside this package. This
should be as thin layer as possible. Ideally just processing CLI params and
displaying results.
"""
from __future__ import absolute_import, unicode_literals

# 3rd party imports
import click

# local imports
import scaffy
from scaffy.config import settings


@click.group()
@click.version_option(version=scaffy.__version__, message='%(version)s')
def cli():
    """

    To get help for a specific command:

        \033[1mscaffy <command> --help\033[0m

    Examples:

        \033[1mscaffy create --help . --help\033[0m

        \033[1mscaffy push --help\033[0m

    """
    settings.load()
    pass



# -*- coding: utf-8 -*-
# pylint: disable=unused-import
""" The application entry point.

Setup completion and import all commands.
"""
from __future__ import absolute_import, unicode_literals

# local imports
from scaffy.commands import cli
__all__ = [
    'cli'
]


try:
    # Only enabled click_completion if psutil package is installed
    import psutil
    import click_completion

    click_completion.init()
except ImportError:
    pass


from scaffy.commands import scaffold    # noqa
from scaffy.commands import local       # noqa
from scaffy.commands import remote      # noqa



# -*- coding: utf-8 -*-
""" scaffy package. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
from os.path import abspath, dirname


__version__ = '0.8.1'
PKG_DIR = abspath(dirname(__file__))



# -*- coding: utf-8 -*-
"""
This is just so that all the configuration is in one place.

Changing the configuration will still require releasing new version.
"""

from __future__ import absolute_import, unicode_literals

# stdlib imports
from os.path import expanduser

# local imports
from scaffy.core.config import Config


class ScaffyConfig(Config):
    """ Configuration for scaffy app. """

    resolution = [
        expanduser('~/.config/scaffy/config.yml'),
        '/etc/scaffy/config.yml',
    ]

    defaults = {
        'CREDENTIALS_PATH': expanduser('~/.config/scaffy/credentials'),
        'BACKENDS': [
            'http://localhost:12001',
            'https://scaffyhub.novocode.net',
        ]
    }


settings = ScaffyConfig()



# -*- coding: utf-8 -*-
""" Commands related directly to scaffolds (create, apply, etc.) """
from __future__ import absolute_import, unicode_literals

from os.path import abspath, basename

# 3rd party imports
import click
from six import string_types

# local imports
from scaffy import scaffold
from scaffy.core import log
from scaffy.core import git
from . import cli


DirPath = click.Path(file_okay=False, dir_okay=True, exists=True)


def marker_def(string):
    """ A simple parameter type to define scaffold markers.

    This will parse a marker in format 'name=value[,other]' to a tuple
    (str, str) of name/value pairs. If there are multiple values, the second
    element in the tuple will be a list of string
    """
    name, value = string.split('=')
    name = name.strip()
    value = value.strip()

    if ',' in value:
        value = value.split(',')

    return name, value


@cli.command('create')
@click.argument('name', type=str, required=True,)
@click.argument('src_dir', type=DirPath)
@click.option(
    '-e',
    '--exclude',
    multiple=True,
    metavar='PATTERN',
    help=('Exclude patterns. To specify multiple patters supply this option '
          'multiple times.')
)
@click.option(
    '-m', '--marker', 'markers',
    type=marker_def,
    multiple=True,
    metavar='NAME=VALUE(S)',
    help=('Markers that will be created in the scaffold. Supply each marker '
          'using -m name=value1,value2 format. In the above example, every '
          "occurrence of either 'value1' or 'value2' will be replaced with "
          "marker named 'name'. The marker will be replaced with real value "
          "when the scaffold is applied (a project is generated)."
          "This also applies to file and directory names.")
)
@click.option(
    '--no-gitignore',
    is_flag=True,
    help='If given, .gitignore will not be read for exclude patterns.'
)
@click.option(
    '-v', '--verbose',
    count=True,
    help="Be more verbose."
)
def create(name, src_dir, markers, exclude, no_gitignore, verbose):
    """ Create new scaffold from a source directory.

    This command will create a new scaffold with the given name from the
    supplied src_dir. A template is just a collection of files, some of them
    with markers. The markers are generated during the scaffold creation (this
    command) and later replaced with real values when the apply command is used.

    The scaffold is created in the local store. To list all scaffolds in the
    local store, use scaffy list. It can also be pushed to scaffy hub so it
    can be accessed from any machine.

    Example:

        \033[1mscaffy create my-scaffold . -e dist -e build \033[0m

    The above call will create a scaffold with name 'my-scaffold' and contents
    of the current directory except for files/dirs matching 'dist' and 'build'
    """
    log.info("Creating scaffold ^35{}".format(name))

    store = scaffold.LocalStore()

    markers = dict(markers)
    markers.setdefault('name', basename(abspath(src_dir)))

    if verbose:
        log.info("Exclude")
        for pattern in sorted(exclude):
            log.info("    ^0{}", pattern)

    if not no_gitignore:
        git_exclude = git.load_gitignore(src_dir)
        exclude = set(exclude) | set(git_exclude) | {'.git'}

        if verbose:
            log.info("Exclude from .gitignore")
            for pattern in sorted(git_exclude):
                log.info("    ^0{}", pattern)

    if name is None:
        name = markers['name']
        if not isinstance(name, string_types):
            name = name[0]      # name is a list of names, pick the first one.

    item = scaffold.Scaffold.create(src_dir, name, exclude, markers)
    store.add(item)


@cli.command('apply')
@click.argument('name', type=str)
@click.argument('project_name', type=str)
@click.option(
    '-o', '--out-path',
    type=DirPath,
    default='.',
    help=("Specify the path where the new project will be created. The"
          "project folder will be created within the given path.")
)
def apply_scaffold(name, project_name, out_path):
    """ Generate project using the given scaffold.

    The project will have all the scaffold contents with 'name' markers replaced
    by project_name. Be default, the project folder will be created inside the
    current working directory. This can be changed using the --out-path option.
    """
    store = scaffold.LocalStore()
    item = store.load(name)
    item.apply(project_name, out_path)
